// APPLY - execute commands on selected files and directories - Version 1.7
// M.Winterhoff, Geschw.-Scholl-Ring 17, 38444 Wolfsburg, Germany, mawin@gmx.net
// http://www.geocities.com/mwinterhoff
// compile using Borland-C++ in large memory model

#include <dos.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <dir.h>
#include <bios.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <stdlib.h>

typedef enum{FALSE,TRUE} BOOL;
typedef struct stamp
{
    BOOL timeflag;
    BOOL secflag;
    unsigned int time;
    BOOL dateflag;
    unsigned int date;
}
STAMP;
char cmdline[128];
char * cmd[16];
int cmds=0;
char * parm[16];
int parms=0;
char exclude[16][81];
int excludes=0;
char newer[16][16];
int newers=0;
char older[16][16];
int olders=0;
char change[16][81];
int changes=0;
long count;
long bytes;
int level=9;
char key;
BOOL query=FALSE;
BOOL yank=FALSE;
BOOL longopt=FALSE;
BOOL recursive=FALSE;
BOOL echo=TRUE;
BOOL dirscan=FALSE;
BOOL uni=FALSE;
BOOL hidden=FALSE;
BOOL arc0=FALSE;
BOOL arc1=FALSE;
BOOL wait=FALSE;
STAMP after,before;

unsigned char upper[]=
    "                                 !\"#$%&\'()*+,-.\\0123456789:; = ?"
    "@ABCDEFGHIJKLMNOPQRSTUVWXYZ \\ ^_`ABCDEFGHIJKLMNOPQRSTUVWXYZ{ }~\x7F"
    "EAAEEEIIIOOUUYAIOU"
    "\xFF";

BOOL match(char *sSource,char *sPattern,BOOL ext)
{
    int i;

    while(*sPattern)
    {
        if(*sPattern=='*')
        {
	    i=strlen(sSource);
            do
            {
                if(match(sSource+i,sPattern+1,ext)) return TRUE;
            }
            while(i-->0);
            return FALSE;
        }
        if(*sPattern=='?'&&*sSource!='\0'&&*sSource!='.'||upper[(unsigned char)*sPattern]==*sSource)
        {
            sPattern++;
            sSource++;
        }
        else if(*sSource=='\0'&&sPattern[0]=='.')
        {
            if(sPattern[1]=='\0') return ext;
            return sPattern[1]=='*'&&sPattern[2]=='\0';
        }
        else
        {
            return FALSE;
        }
    }
    return *sSource=='\0';
}

char * parstime(char p[],unsigned int * t,BOOL * b,BOOL * secflag)
{
    int i,h,m,s;

    i=h=m=s=0;
    while(p[i]==' '||p[i]=='\t') i++;
    if(!isdigit(p[i])) return p;
    h=p[i]-'0'; i++;
    if(isdigit(p[i])) { h=10*h+p[i]-'0'; i++; }
    if(p[i]!=':') return p;
    i++;
    if(!isdigit(p[i])) return p;
    m=p[i]-'0'; i++;
    if(isdigit(p[i])) { m=10*m+p[i]-'0'; i++; }
    if(p[i]==':')
    {
	i++;
	if(!isdigit(p[i])) return p;
	s=p[i]-'0'; i++;
	if(isdigit(p[i])) { s=10*s+p[i]-'0'; i++; }
	*secflag=TRUE;
    }
    if(p[i]=='a'||p[i]=='A')
    {
	if(h>12) return p;
	h=h%12; i++;
	if(p[i]=='m'||p[i]=='M') i++;
    }
    else if(p[i]=='p'||p[i]=='P')
    {
	if(h>12) return p;
	h=h%12+12; i++;
	if(p[i]=='m'||p[i]=='M') i++;
    }
    if(h>23||m>59||s>59) return p;
    *t=2048*h+32*m+s/2;
    *b=TRUE;
    return p+i;
}

char * parsdate(char p[],unsigned int * t,BOOL * b)
{
    int i,d,m,y;

    i=d=m=y=0;
    while(p[i]==' '||p[i]=='\t') i++;
    if(!isdigit(p[i])) return p;
    d=p[i]-'0'; i++;
    if(isdigit(p[i])) { d=10*d+p[i]-'0'; i++; }
    if(p[i]=='/')
    {
	i++;
	m=d;
	if(!isdigit(p[i])) return p;
	d=p[i]-'0'; i++;
	if(isdigit(p[i])) { d=10*d+p[i]-'0'; i++; }
	if(p[i]!='/') return p;
    }
    else if(p[i]=='-')
    {
	i++;
	m=d;
	if(!isdigit(p[i])) return p;
	d=p[i]-'0'; i++;
	if(isdigit(p[i])) { d=10*d+p[i]-'0'; i++; }
	if(p[i]!='-') return p;
    }
    else if(p[i]=='.')
    {
	i++;
	if(!isdigit(p[i])) return p;
	m=p[i]-'0'; i++;
	if(isdigit(p[i])) { m=10*m+p[i]-'0'; i++; }
	if(p[i]!='.') return p;
    }
    else return p;
    i++;
    if(!isdigit(p[i])) return p;
    y=p[i++]-'0';
    while(isdigit(p[i])) y=10*y+p[i++]-'0';
    if(d>31||m>12) return p;
    if(y<80) y+=2000; else if(y<100) y+=1900;
    if(y<1980) return p;
    *t=512*(y-1980)+32*m+d;
    *b=TRUE;
    return p+i;
}

unsigned int translate(char name[])
{
    union REGS reg;
    struct SREGS sreg;
    char path[81];

    reg.x.ax=0x6000;
    sreg.ds=FP_SEG((char far *)name);
    reg.x.si=FP_OFF((char far *)name);
    sreg.es=FP_SEG((char far *)path);
    reg.x.di=FP_OFF((char far *)path);
    intdosx(&reg,&reg,&sreg);
    strcpy(name,path);
    return reg.x.ax;
}

int doit(char filename[],unsigned int ftime,unsigned int fdate)
{
    char drive[3];
    char path[65];
    char name[9];
    char ext[5];
    FILE *f;
    FILE *g;
    char * from;
    char * to;
    char * percent;
    char * ptr;
    unsigned int date,time;
    char command[128];
    int n,retcode,i,j,handle;

    retcode=0;
    if(longopt)
    {
	fnsplit(filename,drive,path,name,ext);
    }
    else
    {
	strcpy(command,filename);
	translate(command);
	fnsplit(command,drive,path,name,ext);
        if(!name[0]) fnsplit(filename,drive,path,name,ext);
    }
    for(n=0;n<newers;n++)
    {
	from=newer[n];
	command[0]='\0';
        for(percent=from;*percent!='\0'&&*percent!='%'&&*percent!='$';percent++) ;
	while(*percent!='\0')
	{
	    if(percent>from)
	    {
		to=strchr(command,'\0');
		memcpy(to,from,percent-from);
		to[percent-from]='\0';
	    }
	    switch(toupper(*++percent))
	    {
	    case 'D':
		strcat(command,drive);
		break;
	    case 'P':
		strcat(command,path);
		break;
	    case 'N':
		strcat(command,name);
		break;
	    case 'E':
		if(ext[0]!='\0') strcat(command,ext+1);
		break;
	    case 'F':
		strcat(command,filename);
		break;
	    case 'R':
		i=strlen(filename);
		while(i>0&&filename[i-1]!='\\'&&filename[i-1]!=':') i--;
		j=strlen(command);
		if(i>0)
		{
		    strcpy(command+j,filename);
		    if(filename[i-1]=='\\')
		    {
			command[j+i-1]='\0';
		    }
		    else
		    {
			command[j+i]='\0';
		    }
		}
		if(i==0||filename[i-1]==':')
		{
		    strcpy(command+j+i,".");
		}
		while(command[j]!='\0')
		{
		    if(command[j]=='\\') command[j]='/';
		    j++;
		}
		break;
	    case 'S':
		strcat(command,name);
		strcat(command,ext);
		break;
	    case 'X':
		strcat(command,drive);
		strcat(command,path);
		strcat(command,name);
		strcat(command,ext);
		break;
	    default:
		to=strchr(command,'\0');
		*to++=*percent;
		*to='\0';
		break;
	    }
	    from=percent+(*percent!='\0');
            for(percent=from;*percent!='\0'&&*percent!='%'&&*percent!='$';percent++) ;
	}
	strcat(command,from);
        if(_dos_open(command,O_RDONLY,&handle)==0)
        {
            _dos_getftime(handle,&date,&time);
            _dos_close(handle);
            if(date>fdate||date==fdate&&time>=ftime) return 1;
        }
        else
        {
            fprintf(stderr,"Can not open %s\n",command);
            return 1;
        }
    }
    for(n=0;n<olders;n++)
    {
	from=older[n];
	command[0]='\0';
        for(percent=from;*percent!='\0'&&*percent!='%'&&*percent!='$';percent++) ;
	while(*percent!='\0')
	{
	    if(percent>from)
	    {
		to=strchr(command,'\0');
		memcpy(to,from,percent-from);
		to[percent-from]='\0';
	    }
	    switch(toupper(*++percent))
	    {
	    case 'D':
		strcat(command,drive);
		break;
	    case 'P':
		strcat(command,path);
		break;
	    case 'N':
		strcat(command,name);
		break;
	    case 'E':
		if(ext[0]!='\0') strcat(command,ext+1);
		break;
	    case 'F':
		strcat(command,filename);
		break;
	    case 'R':
		i=strlen(filename);
		while(i>0&&filename[i-1]!='\\'&&filename[i-1]!=':') i--;
		j=strlen(command);
		if(i>0)
		{
		    strcpy(command+j,filename);
		    if(filename[i-1]=='\\')
		    {
			command[j+i-1]='\0';
		    }
		    else
		    {
			command[j+i]='\0';
		    }
		}
		if(i==0||filename[i-1]==':')
		{
		    strcpy(command+j+i,".");
		}
		while(command[j]!='\0')
		{
		    if(command[j]=='\\') command[j]='/';
		    j++;
		}
		break;
	    case 'S':
		strcat(command,name);
		strcat(command,ext);
		break;
	    case 'X':
		strcat(command,drive);
		strcat(command,path);
		strcat(command,name);
		strcat(command,ext);
		break;
	    default:
		to=strchr(command,'\0');
		*to++=*percent;
		*to='\0';
		break;
	    }
	    from=percent+(*percent!='\0');
	    for(percent=from;*percent!='\0'&&*percent!='%'&&*percent!='$';percent++) ;
	}
	strcat(command,from);
        if(_dos_open(command,O_RDONLY,&handle)==0)
        {
            _dos_getftime(handle,&date,&time);
            _dos_close(handle);
            if(fdate>date||fdate==date&&ftime>=time) return 1;
        }
        else
        {
            fprintf(stderr,"Can not open %s\n",command);
            return 1;
        }
    }
    for(n=0;n<changes;n++)
    {
	from=change[n];
	command[0]='\0';
        for(percent=from;*percent!='\0'&&*percent!='%'&&*percent!='$';percent++) ;
	while(*percent!='\0')
	{
	    if(percent>from)
	    {
		to=strchr(command,'\0');
		memcpy(to,from,percent-from);
		to[percent-from]='\0';
	    }
	    switch(toupper(*++percent))
	    {
	    case 'D':
		strcat(command,drive);
		break;
	    case 'P':
		strcat(command,path);
		break;
	    case 'N':
		strcat(command,name);
		break;
	    case 'E':
		if(ext[0]!='\0') strcat(command,ext+1);
		break;
	    case 'F':
		strcat(command,filename);
		break;
	    case 'R':
		i=strlen(filename);
		while(i>0&&filename[i-1]!='\\'&&filename[i-1]!=':') i--;
		j=strlen(command);
		if(i>0)
		{
		    strcpy(command+j,filename);
		    if(filename[i-1]=='\\')
		    {
			command[j+i-1]='\0';
		    }
		    else
		    {
			command[j+i]='\0';
		    }
		}
		if(i==0||filename[i-1]==':')
		{
		    strcpy(command+j+i,".");
		}
		while(command[j]!='\0')
		{
		    if(command[j]=='\\') command[j]='/';
		    j++;
		}
		break;
	    case 'S':
		strcat(command,name);
		strcat(command,ext);
		break;
	    case 'X':
		strcat(command,drive);
		strcat(command,path);
		strcat(command,name);
		strcat(command,ext);
		break;
	    default:
		to=strchr(command,'\0');
		*to++=*percent;
		*to='\0';
		break;
	    }
	    from=percent+(*percent!='\0');
            for(percent=from;*percent!='\0'&&*percent!='%'&&*percent!='$';percent++) ;
	}
	strcat(command,from);
        f=fopen(command,"rb");
        if(f)
        {
            g=fopen(filename,"rb");
            if(g)
            {
                do
                {
                    i=getc(f);
                    j=getc(g);
                }
                while(i!=-1&&i==j);
                fclose(g);
            }
            else
            {
                fprintf(stderr,"Can not open %s\n",filename);
                fclose(f);
                return 1;
            }
            fclose(f);
            if(i==-1&&j==-1) return 1;
        }
        else
        {
	    fprintf(stderr,"Can not open %s\n",command);
            return 1;
        }
    }
    if(cmds==0)
    {
	printf("%s\n",filename);
    }
    else
    {
        for(n=0;n<cmds&&retcode!=-1&&bioskey(1)!=0x011B;n++)
        {
	    from=cmd[n];
	    command[0]='\0';
            for(percent=from;*percent!='\0'&&*percent!='%'&&*percent!='$';percent++) ;
	    while(*percent!='\0')
	    {
	        to=strchr(command,'\0');
	        if(percent>from)
	        {
		    memcpy(to,from,percent-from);
		    to[percent-from]='\0';
                    to+=percent-from;
	        }
                if(*percent=='%')
                {
                    i=strcspn(++percent,"%");
                    memcpy(to,percent,i);
                    to[i]='\0';
		    ptr=getenv(to);
                    if(ptr) strcpy(to,ptr);
                    percent+=i;
                }
	        else switch(toupper(*++percent))
	        {
	        case 'D':
		    strcat(command,drive);
		    break;
	        case 'P':
		    strcat(command,path);
		    break;
	        case 'N':
		    strcat(command,name);
		    break;
	        case 'E':
		    if(ext[0]!='\0') strcat(command,ext+1);
		    break;
	        case 'F':
		    strcat(command,filename);
		    break;
	        case 'R':
		    i=strlen(filename);
		    while(i>0&&filename[i-1]!='\\'&&filename[i-1]!=':') i--;
		    j=strlen(command);
		    if(i>0)
		    {
		        strcpy(command+j,filename);
		        if(filename[i-1]=='\\')
			{
			    command[j+i-1]='\0';
		        }
		        else
		        {
			    command[j+i]='\0';
		        }
		    }
		    if(i==0||filename[i-1]==':')
		    {
		        strcpy(command+j+i,".");
		    }
		    while(command[j]!='\0')
		    {
		        if(command[j]=='\\') command[j]='/';
		        j++;
		    }
		    break;
	        case 'S':
		    strcat(command,name);
		    strcat(command,ext);
		    break;
	        case 'X':
		    strcat(command,drive);
		    strcat(command,path);
		    strcat(command,name);
		    strcat(command,ext);
		    break;
	        default:
		    to=strchr(command,'\0');
		    *to++=*percent;
		    *to='\0';
		    break;
	        }
	        from=percent+(*percent!='\0');
                for(percent=from;*percent!='\0'&&*percent!='%'&&*percent!='$';percent++) ;
	    }
	    strcat(command,from);
	    if(echo) printf("%s\n",command);
	    if(!uni&&command[0]) retcode=system(command);
        }
    }
    return retcode;
}

void search(char pathname[],int level)
{
    char drive[3];
    char path[65];
    char name[13];
    char ext[5];
    BOOL done;
    int k;
    struct ffblk buffer;

    fnsplit(pathname,drive,path,name,ext);
    strcat(name,ext);
    fnmerge(pathname,drive,path,"*",".*");
    done=findfirst(pathname,&buffer,FA_DIREC|(hidden?FA_HIDDEN:0));
    while(!done&&key!='E'&&key!='\x1B'&&bioskey(1)!=0x011B)
    {
	fnmerge(pathname,drive,path,buffer.ff_name,"");
	if(!(buffer.ff_attrib&FA_DIREC)!=dirscan&&buffer.ff_name[0]!='.')
        {
            if(!match(buffer.ff_name,name,!strchr(buffer.ff_name,'.'))) goto skip;
	    if(arc0&&(buffer.ff_attrib&FA_ARCH)!=0) goto skip;
	    if(arc1&&(buffer.ff_attrib&FA_ARCH)==0) goto skip;
	    for(k=0;k<excludes;k++)
	    {
		if(match(buffer.ff_name,exclude[k],!strchr(buffer.ff_name,'.'))) goto skip;
	    }
	    k=after.secflag?0xFFFF:0xFFF0;
	    if(after.timeflag&&after.dateflag)
	    {
		if(buffer.ff_fdate<after.date||buffer.ff_fdate==after.date&&(buffer.ff_ftime&k)<after.time) goto skip;
	    }
	    else if(after.timeflag)
	    {
		if((buffer.ff_ftime&k)<after.time) goto skip;
	    }
	    else if(after.dateflag)
	    {
		if(buffer.ff_fdate<after.date) goto skip;
	    }
	    k=after.secflag?0xFFFF:0xFFF0;
	    if(before.timeflag&&before.dateflag)
	    {
		if(buffer.ff_fdate>before.date||buffer.ff_fdate==before.date&&(buffer.ff_ftime&k)>before.time) goto skip;
	    }
	    else if(before.timeflag)
	    {
		if((buffer.ff_ftime&k)>before.time) goto skip;
	    }
	    else if(before.dateflag)
	    {
		if(buffer.ff_fdate>before.date) goto skip;
	    }
	    if(longopt) translate(pathname);
	    if(query&&key!='A')
	    {
		printf("%s (Yes/No/All/Esc) ? ",pathname);
		do
		{
		    key=toupper(bioskey(0)&0xFF);
		}
		while(key!='Y'&&key!='J'&&key!='N'&&key!='A'&&key!='E'&&key!='\x1B');
		printf("\n");
	    }
	    if(key=='Y'||key=='J'||key=='A')
	    {
		switch(doit(pathname,buffer.ff_ftime,buffer.ff_fdate))
		{
                case -1:
		    key='E';
                    break;
                case 0:
		    count++;
		    if(!dirscan) bytes+=buffer.ff_fsize;
                    break;
		}
	    }
	}
skip:
	if(buffer.ff_attrib&FA_DIREC)
	{
	    if(recursive&&level>1&&buffer.ff_name[0]!='.')
	    {
		fnmerge(pathname,drive,path,buffer.ff_name,"");
		strcat(pathname,"\\");
		strcat(pathname,name);
		search(pathname,level-1);
	    }
	}
        done=findnext(&buffer);
    }
}

int main(int argc,char *argv[])
{
    int i;
    char * cp;
    char pathname[81];
    time_t t;
    struct tm *now;
    FILE *list;
    FILE *f;
    struct tm last;
    struct stat buf;
    BOOL anyopt;

    movedata(_psp,0x81,FP_SEG((char far *)cmdline),FP_OFF((char far *)cmdline),*(unsigned char far *)MK_FP(_psp,0x80));
    cmdline[*(unsigned char far *)MK_FP(_psp,0x80)]='\0';
    cp=cmdline;
    anyopt=FALSE;
    while(*cp!='\0')
    {
	if(*cp==' '||*cp=='\t')
	{
	    cp++;
	}
	else if(*cp=='-'||*cp=='/')
	{
	    cp++;
	    switch(toupper(*cp++))
	    {
	    case 'A':
		cp=parsdate(cp,&after.date,(BOOL *)&after.dateflag);
		cp=parstime(cp,&after.time,(BOOL*)&after.timeflag,(BOOL *)&after.secflag);
		if(!after.dateflag&&!after.timeflag) arc0=TRUE;
		break;
	    case 'B':
		cp=parsdate(cp,&before.date,(BOOL *)&before.dateflag);
		cp=parstime(cp,&before.time,(BOOL *)&before.timeflag,(BOOL *)&before.secflag);
		if(!before.dateflag&&!before.timeflag) arc1=TRUE;
		break;
            case 'N':
                while(*cp==' '||*cp=='\t') cp++;
                for(i=0;cp[i]!=' '&&cp[i]!='\0'&&cp[i]!='\t';i++)
                {
                    newer[newers][i]=cp[i];
                }
		if(i<1||i>80)
		{
		    fprintf(stderr,"Ungltiger Dateiname bei Option -n\n");
		    return 1;
		}
                newer[newers][i]='\0';
                cp+=i;
		newers++;
		break;
            case 'O':
                while(*cp==' '||*cp=='\t') cp++;
                for(i=0;cp[i]!=' '&&cp[i]!='\0'&&cp[i]!='\t';i++)
                {
                    older[olders][i]=cp[i];
                }
		if(i<1||i>80)
		{
		    fprintf(stderr,"Ungltiger Dateiname bei Option -o\n");
		    return 1;
		}
                older[olders][i]='\0';
                cp+=i;
		olders++;
		break;
            case 'C':
                while(*cp==' '||*cp=='\t') cp++;
                for(i=0;cp[i]!=' '&&cp[i]!='\0'&&cp[i]!='\t';i++)
                {
                    change[changes][i]=cp[i];
                }
		if(i<1||i>80)
		{
		    fprintf(stderr,"Ungltiger Dateiname bei Option -c\n");
		    return 1;
		}
                change[changes][i]='\0';
                cp+=i;
		changes++;
		break;
	    case 'L':
	    case 'V':
		longopt=TRUE;
		break;
	    case 'E':
		echo=FALSE;
		break;
	    case 'U':
		uni=TRUE;
		break;
	    case 'D':
                dirscan=TRUE;
		break;
	    case 'T':
	    case 'R':
	    case 'S':
		recursive=TRUE;
                if(isdigit(*cp))
                {
                    level=*cp-'0';
                    cp++;
                }
		break;
	    case 'P':
	    case 'Q':
		query=TRUE;
		break;
	    case 'H':
		hidden=TRUE;
		break;
	    case 'W':
		wait=TRUE;
		break;
	    case 'X':
                while(*cp==' '||*cp=='\t') cp++;
                for(i=0;cp[i]!=' '&&cp[i]!='\0'&&cp[i]!='\t';i++)
                {
                    exclude[excludes][i]=cp[i];
                }
		if(i<1||i>12)
		{
		    fprintf(stderr,"Ungltiger Dateiname bei Option -x\n");
		    return 1;
		}
                exclude[excludes][i]='\0';
                cp+=i;
		excludes++;
		break;
	    case 'Y':
	        yank=TRUE;
	        break;
	    default:
		fprintf(stderr,"Unknown Option %c\n",*(cp-1));
		return 1;
	    }
            anyopt=TRUE;
	}
	else if(*cp=='\"'||*cp=='\'')
	{
            i=*cp++;
	    if(cmds>16)
	    {
		fprintf(stderr,"Too many commands (max. 16)\n");
		return 1;
	    }
	    cmd[cmds++]=cp;
            while(*cp!=i)
            {
                if(*cp=='\0')
                {
		    fprintf(stderr,"Unterminated command\n");
                    return 1;
                }
                cp++;
            }
	    *cp++='\0';
	}
	else if(parms==16)
	{
	    fprintf(stderr,"Too many parameters (max. 16)\n");
	    return 1;
	}
	else
	{
	    parm[parms++]=cp;
	    do
	    {
		cp++;
	    }
	    while(*cp!='\0'&&*cp!=' '&&*cp!='\t');
	    if(*cp!='\0') *cp++='\0';
	}
    }
    if(parms==0)
    {
	if(anyopt)
	{
	    parm[parms++]="*.*";
	}
	else
	{
	    fprintf(stderr,"%s",
			   "APPLY - execute commands on selected files or directories         - Version 1.8\n"
			   "M.Winterhoff, Geschw.-Scholl-Ring 17, 38444 Wolfsburg, Germany, mawin@gmx.net\n"
			   "http://www.geocities.com/mwinterhoff\n"
			   "usage:    APPLY {[d:][\\path\\]maskname|\"command\"|'command'|options|@listfile}\n\n"
			   "where:    d:           optional drive specifier\n"
			   "          path         optional path name\n"
			   "          maskname     any legal DOS file name, directory name or file mask\n"
			   "          command      any legal DOS command including vars explained below\n"
			   "          listfile     text file containing [d:][\path\]masknames\n\n"
			   "vars:     $d           drive of selected file in form d:\n"
			   "          $r           relative path of selected file\n"
			   "          $p           path of selected file in form \\path\\\n"
			   "          $n           name of selected file without extension\n"
			   "          $e           extension of selected file without dot\n"
			   "          $f           pathname of selected file as specified\n"
			   "          $l           long pathname of selected file\n"
			   "          $s           name and extension of selected file\n"
			   "          $%           to represent a % in command\n"
			   "          $$           to represent a $ in command\n"
			   "          $\"           to represent a \" in command\n"
			   "          %envvar%     any enviroment variable envvar replaced\n\n"
			   "Press [Enter] to continue explanation, [Esc] to abort\r");
	    if(bioskey(0)==0x011B) return 0;
	    fprintf(stderr,"%s",
			   "options:  /a           match only files not backed up  \n"
			   "          /b           match only files backed up\n"
			   "          /a date      match only files after and including date\n"
			   "          /b date      match only files before and including date\n"
			   "          /a time      match only files after and including time of every day\n"
			   "          /b time      match only files before and including time of every day\n"
			   "          /a date time match only files after and including date&time\n"
		           "          /b date time match only files before and including date&time\n"
			   "          /c file      only execute if file differs\n"
                           "          /d           operate on directory names instead of filenames\n"
		           "          /e           do not echo commands before execution\n"
		           "          /u           do not execute commands (just echo)\n"
		           "          /m file      execute if file is missing\n"
                           "          /n file      do not execute if file is newer\n"
                           "          /o file      do not execute if file is older\n"
		           "          /p /q        query each file\n"
			   "          /l /v        display full pathname of each file processed\n"
		           "          /s /t /r[n]  recursivly scan subdirs [up to n levels deep]\n"
		           "          /h           include hidden files also\n"
		           "          /w           wait for keypress before proceeding\n"
		           "          /x maskname  exclude files matching maskname\n"
		           "          /y           match only files changed after last execution\n\n"
                           "Press [Enter] to continue explanation, [Esc] to abort\r");
            if(bioskey(0)==0x011B) return 0;
	    fprintf(stderr,"%s",
		           "format:   time         hh:mm[:ss] or hh:mm[:ss](a|p)[m]\n"
		           "          date         dd.mm.[yy]yy or mm/dd/[yy]yy or mm-dd-[yy]yy\n"
		           "          file         any character string with vars replaced\n\n"
                           "maskname: *            will match any filename with or without extension\n"
                           "          *.*          will match any filename with or without extension\n"
                           "          ????         will match any filename with exactly 4 characters\n"
			   "          *.           will match only filenames without extension\n"
                           "          *.?*         will match only filenames with extension\n"
                           "          *A*.*B*      will match files containing A in name, B in extension\n\n"
			   "samples:  APPLY *.BAK \"del $f\"                          \n"
		           "          will delete any BAK-files in current directory like del *.BAK\n"
		           "          APPLY C:\*.SWP /s \"move $f A:\\$s\"\n"
			   "          will move any *.SWP-files from any dir on drive A: to root dir\n"
		           "          APPLY /S /W *.TXT *.WRI *.DOC \"COPY $f A:\" /Q\n"
		           "          will ask for any document file to copy on drive A: after disk change\n"
		           "          APPLY *.C \"del $n.o\" \"ren $n.obj $n.o\" \"cl $f\" /X noname*.c\n"
		           "          will recompile any C-file after saving object file but not noname*.c\n"
		           "          APPLY *E*X* \"del $f\"\n"
		           "          will delete any file containing E folloed by X\n\n"
			   "This program is freeware. Use at your own risk. No part of it may be used\n"
			   "commercially. No fees may be charged on copying. Always distribute with source.\n");
	    return 1;
	}
    }
    if(yank)
    {
	time(&t);
	now=localtime(&t);
	strcpy(strrchr(argv[0],'.'),".CFG");
	f=fopen(argv[0],"rb");
	if(f)
	{
	    fread(&last,sizeof(struct tm),1,f);
	    fclose(f);
	}
	else
	{
	    memcpy(&last,now,sizeof(struct tm));
	}
	if(echo) fprintf(stderr,"apply to files updated after %d.%2d.%d %d:%02d:%02d\n",last.tm_mday,last.tm_mon+1,last.tm_year+1900,last.tm_hour,last.tm_min,last.tm_sec);
        after.timeflag=TRUE;
	after.secflag=TRUE;
        after.time=2048*last.tm_hour+32*last.tm_min+last.tm_sec/2;
        after.dateflag=TRUE;
        after.date=512*(last.tm_year-80)+32*(last.tm_mon+1)+last.tm_mday;
    }
    if(wait)
    {
	fprintf(stderr,"Press [Enter] to continue, [Esc] to abort...\n");
	if(bioskey(0)==0x011B) return 3;
    }
    count=0L;
    bytes=0L;
    key='Y';
    for(i=0;i<parms&&bioskey(1)!=0x011B;i++)
    {
	strcpy(pathname,parm[i]);
	if(pathname[0]=='@')
	{
	    list=fopen(pathname+1,"rt");
	    if(!list)
	    {
		fprintf(stderr,"Can not open list file %s",pathname+1);
	    }
	    else
	    {
		while(fscanf(list,"%s",pathname))
		{
                    if(!dirscan)
                    {
			if(*(strchr(pathname,'\0')-1)=='\\'||*(strchr(pathname,'\0')-1)==':')
		        {
			    strcat(pathname,"*.*");
		        }
		        else if(strchr(pathname,'?')==NULL&&strchr(pathname,'*')==0&&stat(pathname,&buf)==0&&(buf.st_mode&S_IFDIR)!=0)
		        {
			    strcat(pathname,"\\*.*");
		        }
                    }
		    search(pathname,level);
		}
		fclose(list);
	    }
	}
	else
	{
            if(!dirscan)
            {
	        if(*(strchr(pathname,'\0')-1)=='\\'||*(strchr(pathname,'\0')-1)==':')
	        {
		    strcat(pathname,"*.*");
	        }
	        else if(strchr(pathname,'?')==NULL&&strchr(pathname,'*')==0&&stat(pathname,&buf)==0&&(buf.st_mode&S_IFDIR)!=0)
		{
		    strcat(pathname,"\\*.*");
	        }
            }
	    search(pathname,level);
	}
    }
    if(bioskey(1)==0x011B) bioskey(0);
    if(dirscan)
    {
        fprintf(stderr,"%14ld director%s\n",count,count!=1?"ies":"y",bytes);
    }
    else
    {
        fprintf(stderr,"%14ld file%s, %ld bytes\n",count,count!=1?"s":"",bytes);
    }
    if(yank)
    {
	f=fopen(argv[0],"wb");
	if(f)
	{
	    time(&t);
	    now=localtime(&t);
	    fwrite(now,sizeof(struct tm),1,f);
	    if(echo) fprintf(stderr,"time updated to %d.%2d.%d %d:%02d:%02d\n",now->tm_mday,now->tm_mon+1,now->tm_year+1900,now->tm_hour,now->tm_min,now->tm_sec);
	    fclose(f);
	}
    }
    return 0;
}
